#!/usr/bin/env python
"""


Module to perform structure tensor analysis
"""
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors
from scipy import ndimage
from os.path import join


def structure_tensor(im, sigma):
    # Assumes IM is a 2D numpy array (double precision)

    if im.max() > 1:
        im = im / im.max()

    # Calculate directional derivatives dIM/dx and dIM/dy
    # imx,imy = np.gradient(im)
    imx = ndimage.sobel(im, axis=1, mode='mirror')
    imy = ndimage.sobel(im, axis=0, mode='mirror')

    # Calculate orientation for each pixel
    # Compononents of the symmetric structure tensor:
    # [d^2/dx       d/dx*d/dy]
    # [d/dx*d/dy    d^2/dy]
    xy = ndimage.gaussian_filter(imy * imx, sigma=sigma, mode='mirror')
    xx = ndimage.gaussian_filter(imx ** 2, sigma=sigma, mode='mirror')
    yy = ndimage.gaussian_filter(imy ** 2, sigma=sigma, mode='mirror')

    # Orientation map
    orientation = 0.5 * np.arctan2(2 * xy, (yy - xx))
    # Energy map
    energy = xx + yy

    return orientation, energy


"""Orientation conversion"""


def convert_rad_to_deg(O):
    return (O / (np.pi) * 180)


def wrap(O):
    return np.arctan2(np.sin(O), np.cos(O))


"""Visualization"""


def save_intermediate(idx, map, savepath):

    if idx in [50, 100, 150, 200, 500, 750]:

        cmap_settings = ('jet', None)  # build 1st colour map
        symmmetric_jet = symmetrical_colormap(cmap_settings=cmap_settings, new_name=None)
        symmmetric_jet.set_bad(color='black')  # set nans to black background

        brb = build_custom_continuous_cmap([0, 0, 255], [255, 0, 0],
                                           [0, 0, 255])  # build 2nd colour map red = left/right, blue = up/down
        brb.set_bad(color='black')  # set nans to black background

        plt.figure(figsize=(20, 20))
        plt.imshow(map, cmap=symmmetric_jet, vmax=90, vmin=-90)
        plt.savefig(join(savepath, "orientation_{}_jet.png".format(idx)))
        plt.close()

        plt.figure(figsize=(20, 20))
        plt.imshow(map, cmap=brb, vmax=90, vmin=-90)
        plt.savefig(join(savepath, "orientation_{}_brb.png".format(idx)))
        plt.colorbar()
        plt.close()


"""Colour maps"""


def symmetrical_colormap(cmap_settings, new_name=None):
    ''' This function take a colormap and create a new one, as the concatenation of itself by a symmetrical fold.
    '''
    # get the colormap
    cmap = plt.cm.get_cmap(*cmap_settings)
    if not new_name:
        new_name = "sym_" + cmap_settings[0]  # ex: 'sym_Blues'

    # this defined the roughness of the colormap, 128 fine
    n = 128

    # get the list of color from colormap
    colors_r = cmap(np.linspace(0, 1, n))  # take the standard colormap # 'right-part'
    colors_l = colors_r[::-1]  # take the first list of color and flip the order # "left-part"

    # combine them and build a new colormap
    colors = np.vstack((colors_l, colors_r))
    mymap = mcolors.LinearSegmentedColormap.from_list(new_name, colors)

    return mymap


def build_custom_continuous_cmap(*rgb_list):
    '''
    Generating any custom continuous colormap, user should supply a list of (R,G,B) color taking the value from [0,255], because this is
    the format the adobe color will output for you.

    Examples::

        test_cmap = build_custom_continuous_cmap([64,57,144],[112,198,162],[230,241,146],[253,219,127],[244,109,69],[169,23,69])
        fig,ax = plt.subplots()
        fig.colorbar(cm.ScalarMappable(norm=colors.Normalize(),cmap=diverge_cmap),ax=ax)

    .. image:: ./_static/custom_continuous_cmap.png
        :height: 400px
        :width: 550px
        :align: center
        :target: target

    '''
    all_red = []
    all_green = []
    all_blue = []
    for rgb in rgb_list:
        all_red.append(rgb[0])
        all_green.append(rgb[1])
        all_blue.append(rgb[2])
    # build each section
    n_section = len(all_red) - 1
    red = tuple([(1 / n_section * i, inter_from_256(v), inter_from_256(v)) for i, v in enumerate(all_red)])
    green = tuple([(1 / n_section * i, inter_from_256(v), inter_from_256(v)) for i, v in enumerate(all_green)])
    blue = tuple([(1 / n_section * i, inter_from_256(v), inter_from_256(v)) for i, v in enumerate(all_blue)])
    cdict = {'red': red, 'green': green, 'blue': blue}
    new_cmap = colors.LinearSegmentedColormap('new_cmap', segmentdata=cdict)
    return new_cmap


def inter_from_256(x):
    return np.interp(x=x, xp=[0, 255], fp=[0, 1])
